"""
Implements the DeviceManager for physical and virtual devices. Every
physical device (a router) has a set of virtual subdevices.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from .processor import Description
from .utils import get_xml_root


class DeviceManager:
    """
    Knows all data about the device and the sub-devices, including the
    available services. Takes an optional `timeout` parameter to limit
    the time waiting for a router response. The optional parameter
    `session` is a reusable connection and can speed up the
    communication with the device. In case `session` is given, `timeout`
    will not get used.
    """

    def __init__(self, timeout=None, session=None):
        self.descriptions = []
        self.services = {}
        self.timeout = timeout
        self.session = session

    @property
    def modelname(self):
        """
        Take the root-device of the first description and return the
        according modelname. This is the name of the Fritz!Box itself.
        Will raise an IndexError if the method is called before
        descriptions are added.
        """
        return self.descriptions[0].device_model_name

    @property
    def system_version(self):
        """
        Returns the system-version as string with minor- and
        patch-level. This corresponds to the OS version reported by the
        router web-interface. Returns None if no system information is
        available.
        """
        version = None
        for description in self.descriptions:
            version = description.system_version
            if version:
                return version
        return None

    @property
    def system_info(self):
        """
        Returns a tuple with Hardwarecode, Major-, Minor-, Patch-Level,
        Buildnumber and Display-String, in this order.
        Return None if this information is not available.
        """
        system_info = None
        for description in self.descriptions:
            system_info = description.system_info
            if system_info and system_info[0]:
                return system_info
        return None

    def add_description(self, source):
        """
        Adds description data about the devices and the according
        services. 'source' is a string with the xml-data, like the
        content of an igddesc- or tr64desc-file.
        """
        root = get_xml_root(source, timeout=self.timeout, session=self.session)
        self.descriptions.append(Description(root))

    def scan(self):
        """
        Scans all available services defined by the description files.
        Must get called after all xml-descriptions are added.
        """
        for description in self.descriptions:
            self.services.update(description.services)

    def load_service_descriptions(self, address, port):
        """
        Triggers the load of the scpd files of the services, so they
        known their actions.
        """
        for service in self.services.values():
            service.load_scpd(
                address, port, timeout=self.timeout, session=self.session
            )

    def serialize(self):
        """
        Returns a json-serializable list with Python-datastructures
        representing the known api of the device.
        """
        return [description.serialize() for description in self.descriptions]

    def deserialize(self, data):
        """
        Fills the collections `self.descriptions` and `self.services`
        with the content provided by `data` (from a json-source).
        """
        for description_data in data:
            self.descriptions.append(Description.from_data(description_data))



"""
Abstract base class for all library classes.

An internal module providing the `AbstractLibraryBase` class. This is an
abstract class that should not get instantiated but should serve as a
base class for library classes providing a common initialisation.

"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from __future__ import annotations

from ..core.fritzconnection import FritzConnection


class AbstractLibraryBase:
    """
    Abstract base class for library classes. Implements the common
    initialisation. The first argument `fc` can be a FritzConnection
    instance. If this argument is given no further arguments are needed.
    If the argument `fc` is not given, all other arguments are forwarded
    to get a FritzConnection instance. These arguments have the same
    meaning as for `FritzConnection.__init__()`. Using positional
    arguments is strongly discouraged. Use keyword arguments instead.
    """
    def __init__(
        self,
        fc: FritzConnection | None = None,
        *args,
        **kwargs
    ):
        if fc is None:
            fc = FritzConnection(*args, **kwargs)
        self.fc = fc

    @property
    def modelname(self) -> str:
        """
        The device modelname. Every library module derived from
        `AbstractLibraryBase` inherits this property.
        """
        return self.fc.modelname



"""
Module to access lists of recent phone calls: incoming, outgoing and
missed ones.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from __future__ import annotations

import datetime

from ..core.processor import (
    processor,
    process_node,
    InstanceAttributeFactory,
    Storage,
)
from ..core.utils import get_xml_root
from .fritzbase import AbstractLibraryBase



__all__ = ['FritzCall', 'Call']


ALL_CALL_TYPES = 0
RECEIVED_CALL_TYPE = 1
MISSED_CALL_TYPE = 2
OUT_CALL_TYPE = 3
ACTIVE_RECEIVED_CALL_TYPE = 9
REJECTED_CALL_TYPE = 10
ACTIVE_OUT_CALL_TYPE = 11

SERVICE = 'X_AVM-DE_OnTel1'


def datetime_converter(date_string):
    if not date_string:
        return date_string
    return datetime.datetime.strptime(date_string, '%d.%m.%y %H:%M')


def timedelta_converter(duration_string):
    if not duration_string:
        return duration_string
    hours, minutes = [int(part) for part in duration_string.split(':', 1)]
    return datetime.timedelta(hours=hours, minutes=minutes)


class FritzCall(AbstractLibraryBase):
    """
    Can dial phone numbers and gives access to lists of recent phone
    calls: incoming, outgoing and missed ones. All parameters are
    optional. If given, they have the following meaning: `fc` is an
    instance of FritzConnection, `address` the ip of the Fritz!Box,
    `port` the port to connect to, `user` the username, `password` the
    password, `timeout` a timeout as floating point number in seconds,
    `use_tls` a boolean indicating to use TLS (default False).
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.calls = None  # Instance of CallCollection to store Call-instances

    def _update_calls(self, num=None, days=None):
        result = self.fc.call_action(SERVICE, 'GetCallList')
        url = result['NewCallListURL']
        if days:
            url += f'&days={days}'
        elif num:
            url += f'&max={num}'
        root = get_xml_root(url, session=self.fc.session)
        self.calls = CallCollection(root)

    def get_calls(
        self,
        calltype: int = ALL_CALL_TYPES,
        update: bool = True,
        num: int | None = None,
        days: int | None = None
    ) -> list[Call]:
        """
        Return a list of Call instances of type calltypes. If calltype
        is 0 all calls are listed. If *update* is True, all calls are
        reread from the router. *num* maximum number of entries in call
        list. *days* number of days to look back for calls e.g. 1: calls
        from today and yesterday, 7: calls from the complete last week.
        """
        if not self.calls:
            update = True
        if update:
            self._update_calls(num, days)
        if calltype == ALL_CALL_TYPES:
            return self.calls.calls
        return [call for call in self.calls if call.type == calltype]

    def get_received_calls(
        self,
        update: bool = True,
        num: int | None = None,
        days: int | None = None
    ) -> list[Call]:
        """
        Return a list of Call instances of received calls. If *update*
        is True, all calls are reread from the router. *num*: maximum
        number of entries in call list. *days*: number of days to look
        back for calls e.g. 1: calls from today and yesterday, 7: calls
        from the complete last week.
        """
        return self.get_calls(RECEIVED_CALL_TYPE, update, num, days)

    def get_missed_calls(
        self,
        update: bool = True,
        num: int | None = None,
        days: int | None = None
    ) -> list[Call]:
        """
        Return a list of Call instances of missed calls. If *update* is
        True, all calls are reread from the router. *num* maximum number
        of entries in call list. *days* number of days to look back for
        calls e.g. 1: calls from today and yesterday, 7: calls from the
        complete last week.
        """
        return self.get_calls(MISSED_CALL_TYPE, update, num, days)

    def get_out_calls(
        self,
        update: bool = True,
        num: int | None = None,
        days: int | None = None
    ) -> list[Call]:
        """
        Return a list of Call instances of outgoing calls. If *update*
        is True, all calls are reread from the router. *num* maximum
        number of entries in call list. *days* number of days to look
        back for calls e.g. 1: calls from today and yesterday, 7: calls
        from the complete last week.
        """
        return self.get_calls(OUT_CALL_TYPE, update, num, days)

    def dial(self, number: str) -> None:
        """
        Dials the given *number* (number must be a string, as phone
        numbers are allowed to start with leading zeros). This method
        has no return value, but will raise an error reported from the
        Fritz!Box on failure. **Note:** The dial-help of the Fritz!Box
        must be activated to make this work.
        """
        arg = {'NewX_AVM-DE_PhoneNumber': number}
        self.fc.call_action('X_VoIP1', 'X_AVM-DE_DialNumber', arguments=arg)

    def hangup(self) -> None:
        """
        Terminates the dialling process initiated by the dial function.
        """
        self.fc.call_action('X_VoIP1', 'X_AVM-DE_DialHangup')


class AttributeConverter:
    """
    Data descriptor returning converted attribute values.
    """
    def __init__(self, attribute_name, converter=str):
        self.attribute_name = attribute_name
        self.converter = converter

    def __set__(self, obj, value):
        return NotImplemented

    def __get__(self, obj, objtype):
        attr = getattr(obj, self.attribute_name)
        try:
            attr = self.converter(attr)
        except (TypeError, ValueError):
            pass
        return attr


@processor
class Call:
    """
    Represents a call with the attributes provided by AVM. Instance
    attributes are *Id*, *Type*, *Called*, *Caller*, *CallerNumber*,
    *CalledNumber*, *Name*, *Device*, *Port*, *Date*, *Duration*,
    *Count* and *Path* which can be set in case of an optional phone
    message. The spelling represents the original xml-node names.
    Additionally, the following attributes can be accessed by lowercase
    names: *id* returning the Id as integer, *type* returning the Type
    as integer, *date* returning the Date as datetime-instance,
    *duration* returning the Duration as timedelta-instance.
    """
    id = AttributeConverter('Id', int)
    type = AttributeConverter('Type', int)
    date = AttributeConverter('Date', datetime_converter)
    duration = AttributeConverter('Duration', timedelta_converter)

    def __init__(self):
        self.Id = None
        self.Type = None
        self.Called = None
        self.Caller = None
        self.CallerNumber = None
        self.CalledNumber = None
        self.Name = None
        self.Device = None
        self.Port = None
        self.Date = None
        self.Duration = None
        self.Count = None
        self.Path = None

    def __str__(self):
        number = self.Called if self.type == 3 else self.Caller
        duration = self.Duration if self.type != 2 else "-"
        if not number:
            number = "-"
        return f'{self.Type:>6}   {number:24}{self.Date:>18}{duration:>12}'


class CallCollection(Storage):
    """
    Container for a sequence of Call instances.
    """
    Call = InstanceAttributeFactory(Call)

    def __init__(self, root):
        self.timestamp = None
        self.calls = list()
        super().__init__(self.calls)
        process_node(self, root)

    def __iter__(self):
        return iter(self.calls)



"""
Module for read-only access to the contents of the Fritz!Box phonebooks.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Authors: Klaus Bremer, David M. Straub


from __future__ import annotations

from warnings import warn

from ..core.processor import (
    processor,
    process_node,
    InstanceAttributeFactory,
    Storage,
    ValueSequencer,
)
from ..core.utils import get_xml_root
from .fritzbase import AbstractLibraryBase


__all__ = ['FritzPhonebook']


SERVICE = 'X_AVM-DE_OnTel1'


class FritzPhonebook(AbstractLibraryBase):
    """
    Interface to access the Fritz!Box phonebooks. All parameters are
    optional. If given, they have the following meaning: `fc` is an
    instance of FritzConnection, `address` the ip of the Fritz!Box,
    `port` the port to connect to, `user` the username, `password` the
    password, `timeout` a timeout as floating point number in seconds,
    `use_tls` a boolean indicating to use TLS (default False).
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.phonebook = None

    def _action(self, actionname, **kwargs):
        return self.fc.call_action(SERVICE, actionname, **kwargs)

    @property
    def phonebook_ids(self) -> list[int]:
        """
        List of integers identifying the phonebooks. This property is
        defined as `phonebook_ids` and as `list_phonebooks` for backward
        compatibility. The property `list_phonebooks` is deprecated and
        may get removed in the future.
        """
        result = self._action('GetPhonebookList')
        try:
            res = result['NewPhonebookList'].split(',')
            res = [int(x) for x in res]
        except KeyError:
            return []
        return res

    # legathy api name for backward compatibility
    def list_phonebooks(self) -> list[int]:
        """
        .. deprecated:: 1.13.0
           Use :func:`phonebook_ids` instead.
        """
        warn('This method is deprecated. Use "phonebook_ids" instead.', DeprecationWarning)
        return self.phonebook_ids

    def phonebook_info(self, id: int) -> dict:
        """
        Get the `name`, `url` and an optional `extra id` of the
        phonebook with integer `id`. Returns a dictionary with the keys
        `name`, `url` and `xid`.
        """
        result = self._action('GetPhonebook', NewPhonebookId=id)
        return {
            'name': result.get('NewPhonebookName'),
            'url': result.get('NewPhonebookURL'),
            'xid': result.get('NewPhonebookExtraID')
        }

    def get_all_name_numbers(self, id: int) -> list[tuple]:
        """
        Returns all entries from the phonebook with the given id as a
        list of tuples. The first item of every tuple is the contact
        name and the second item is the list of numbers for this
        contact.
        """
        url = self.phonebook_info(id)['url']
        self._read_phonebook(url)
        return [
            (contact.name, contact.numbers)
            for contact in self.phonebook.contacts
        ]

    def get_all_names(self, id: int) -> dict:
        """
        Get a dictionary with all names and their phone numbers for the
        phonebook with `id`. If a name is given more than once in a
        single phonebook, the last entry will overwrite the previous
        entries. (That's because the names are the keys in the
        dictionary returned from this method. In this case use the
        method 'get_all_name_numbers()' to get all entries as a list of
        tuples.)
        """
        return {name: number for name, number in self.get_all_name_numbers(id)}

    def get_all_numbers(self, id: int) -> dict:
        """
        Get a dictionary with all phone numbers and the according names
        for the phonebook with `id`. This method is based on the method
        'get_all_names()' and has the same limitations.
        """
        reverse_contacts = dict()
        for name, numbers in self.get_all_names(id).items():
            for number in numbers:
                reverse_contacts[number] = name
        return reverse_contacts

    def lookup_numbers(self, id: int, name: str) -> list[str]:
        """
        Look up the phone numbers of contact `name` in the phonebook
        with `id`. Returns a list of numbers. Will raise a KeyError if
        the name is unknown.
        """
        return self.get_all_names(id)[name]

    def lookup_names(self, id: int, number: str) -> str:
        """
        Look up the names of the contacts with phone number `number` in
        the phonebook with `id`. Will raise a KeyError if the number is
        unknown.
        """
        return self.get_all_numbers(id)[number]

    def _read_phonebook(self, url):
        """
        Read the content of the phonebook with the given `url`. This
        method sets the phone book instance attribute and has no return
        value.
        """
        root = get_xml_root(url, session=self.fc.session)
        self.phonebook = Phonebook()
        process_node(self, root)


@processor
class Services:
    """
    Services container. So far just for an associated email-address.
    """
    def __init__(self):
        self.email = None


@processor
class Person:
    """
    Data storage for a contact name and an image.
    """
    def __init__(self):
        self.realName = None
        self.imageURL = None


@processor
class Telephony:
    """
    Data storage for the phone numbers of the contact and services.
    """
    number = ValueSequencer('numbers')

    def __init__(self):
        self.numbers = list()
        self.services = Services()


@processor
class Contact:
    """
    Represents a contact with `category`- and `uniqueid`-attributes as
    well as `person`- and telephony-sub-nodes.
    """

    def __init__(self):
        self.category = None
        self.uniqueid = None
        self.person = Person()
        self.telephony = Telephony()

    @property
    def name(self):
        return self.person.realName

    @property
    def numbers(self):
        return self.telephony.numbers


@processor
class Phonebook(Storage):
    """
    Represents a phonebook with a timestamp indicating the last
    modification and a list of Contact instances.
    """
    contact = InstanceAttributeFactory(Contact)

    def __init__(self):
        self.timestamp = None
        self.contacts = list()
        super().__init__(self.contacts)



"""
Module to access home-automation devices by the TR-064 and the
AHA-HTTP-Interface.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from __future__ import annotations

import datetime
import itertools
from warnings import warn
from xml.etree import ElementTree as etree

from .fritzbase import AbstractLibraryBase


# tr-064 homeautomation service
SERVICE = 'X_AVM-DE_Homeauto1'

# Constants describing the possible capabilities of a device.
# Values are bit positions in the AVM-FunctionBitMask:
HAN_FUN_UNIT_1 = 0
UNKNOWN_1= 1
LIGHT_BULB = 2
UNKNOWN_2 = 3
ALARM_SENSOR = 4
AVM_BUTTON = 5
RADIATOR_CONTROL = 6
ENERGY_SENSOR = 7
TEMPERATURE_SENSOR = 8
PLUGGABLE = 9
AVM_DECT_REPEATER = 10
MICROPHONE = 11
UNKNOWN_3 = 12
HAN_FUN_UNIT_2 = 13
UNKNOWN_4 = 14
SWITCHABLE = 15
ADJUSTABLE = 16
COLOR_BULB = 17
BLIND = 18
UNKNOWN_5 = 19
HUMIDITY_SENSOR = 20

# offset for removing the prefix "New" from the tr-064 avm-keys
KEY_OFFSET = len("New")


class FritzHomeAutomation(AbstractLibraryBase):
    """
    Interface for fritzbox homeauto service. All parameters are
    optional. If given, they have the following meaning: `fc` is an
    instance of FritzConnection, `address` the ip of the Fritz!Box,
    `port` the port to connect to, `user` the username, `password` the
    password, `timeout` a timeout as floating point number in seconds,
    `use_tls` a boolean indicating to use TLS (default False).
    """

    def _action(self, actionname, *, arguments=None, **kwargs):
        if arguments is None:
            arguments = kwargs
        return self.fc.call_action(SERVICE, actionname, arguments=arguments)

    @property
    def get_info(self) -> dict:
        """
        Return a dictionary with a single key-value pair:
        'NewAllowedCharsAIN': string with all allowed chars for state
        variable AIN
        """
        return self._action('GetInfo')

    def get_device_information_by_index(
        self,
        index: int
    ) -> dict:
        """
        Return a dictionary with all device arguments according to the
        AVM documentation (x_homeauto) at the given internal index.
        Raise a FritzArrayIndexError (subclass of IndexError) on invalid
        index values.
        """
        return self._action('GetGenericDeviceInfos', NewIndex=index)

    def get_device_information_by_identifier(
        self,
        identifier: str
    ) -> dict:
        """
        Returns a dictionary with all device arguments according to the
        AVM documentation (x_homeauto) with the given identifier (AIN).
        Raise an FritzArgumentError on invalid identifier.
        """
        return self._action('GetSpecificDeviceInfos', NewAIN=identifier)

    def device_information(self) -> list[dict]:
        """
        .. deprecated:: 1.12.0
           Use :func:`get_device_information_list` instead.
        """
        warn('This method is deprecated. Use "get_device_information_list" instead.', DeprecationWarning)
        return self.get_device_information_list()

    def get_device_information_list(self) -> list[dict]:
        """
        Returns a list of dictionaries for all known homeauto-devices.
        """
        info = list()
        for n in itertools.count():
            try:
                device_information = self.get_device_information_by_index(n)
            except IndexError:
                break
            info.append(device_information)
        return info

    def get_homeautomation_device(
        self,
        identifier: str | None = None,
        index: int | None = None
    ) -> HomeAutomationDevice | None:
        """
        Returns a HomeAutomationDevice instance. The device can be
        identified by the `identifier` (ain) or the `index` in the
        internal router device list. If both arguments are given,
        `identifier` takes preference. If neither `identifier` nor
        `index` are given `None` gets returned.
        """
        if identifier:
            information = self.get_device_information_by_identifier(identifier)
        elif index:
            information = self.get_device_information_by_index(index)
        else:
            return None
        return HomeAutomationDevice(self, information, identifier)

    def get_homeautomation_devices(self) -> list[HomeAutomationDevice]:
        """
        Returns a list with HomeAutomationDevice instances of all known
        home-automation devices. The list is ordered in the way the
        router provided the data.
        """
        return [
            HomeAutomationDevice(self, information) for information
            in self.get_device_information_list()
        ]

    def set_switch(self, identifier: str, on: bool = True) -> None:
        """
        Sets a switch state on devices providing a switch state.
        'identifier' must be the AIN of the device. 'on' is a boolean
        whether the switch should be on (True) or off (False).
        This method has no return value.
        Raise a FritzArgumentError on invalid identifier.
        """
        arguments = {
            'NewAIN': identifier,
            'NewSwitchState': 'ON' if on else 'OFF'
        }
        self._action('SetSwitch', arguments=arguments)


class HomeAutomationDevice:
    """
    Represents a device for homeautomation.

    `fh` is a FritzHomeAutomation instance. `device_information` is a
    dictionary like the ones returned from FritzHomeAutomation methods
    'get_device_information_by_index()',
    'get_device_information_by_identifier()' or an item from the list of
    dictionaries returned from 'get_device_information_list()'. This
    dictionary returned from the
    'get_device_information_by_identifier()' has no 'NewAIN' entry. In
    this case the argument `identifier` must be provided with the
    missing 'ain'. If both `identifier` is provided and
    `device_information` has a 'NewAIN' entry the latter takes
    preference.

    Instances will have the folloing dynamic created attributes:

        AIN
        DeviceId
        FunctionBitMask
        FirmwareVersion
        Manufacturer
        ProductName
        DeviceName
        Present
        MultimeterIsEnabled
        MultimeterIsValid
        MultimeterPower
        MultimeterEnergy
        TemperatureIsEnabled
        TemperatureIsValid
        TemperatureCelsius
        TemperatureOffset
        SwitchIsEnabled
        SwitchIsValid
        SwitchState
        SwitchMode
        SwitchLock
        HkrIsEnabled
        HkrIsValid
        HkrIsTemperature
        HkrSetVentilStatus
        HkrSetTemperature
        HkrReduceVentilStatus
        HkrReduceTemperature
        HkrComfortVentilStatus
        HkrComfortTemperature

    Depending on the device not all attributes will have a meaning.
    """
    def __init__(
        self,
        fh: FritzHomeAutomation,
        device_information: dict,
        identifier: str | None = None
    ):
        self.fh = fh
        self.AIN = identifier
        self._extraxt_device_information_as_attributes(device_information)

    def __repr__(self):
        """
        Provide some basic information about the device.
        """
        return f"ain: {self.AIN}, {self.Manufacturer} - {self.ProductName}"

    def _extraxt_device_information_as_attributes(self, device_information):
        """
        Takes the device_information, which is a dictionary returned
        from a call like
        'FritzHomeAutomation.get_device_information_by_index()' and
        updates the instance-attributes with the key-value pairs of this
        dictionary. The key-names are changed by stripping the prefix
        'New'. All keys updated by this automatic process are in
        MixedCase style - as provided from the router - while all other
        attributes are snake_case.
        """
        for key, value in device_information.items():
            self.__dict__[key[KEY_OFFSET:]] = value

    def _bitmatch(self, value: int) -> bool:
        """
        Returns a boolean whether the `value` bit is set in
        `self.FunctionBitMask`.
        """
        feature_bit = 1 << value
        return feature_bit & self.FunctionBitMask == feature_bit  # type: ignore

    def call_http(self, command: str, **kwargs) -> dict[str, str]:
        """
        Shortcut to access the http-interface of the router.

        Used to send a `command` with the given keyword-arguments to
        _this_ device. Raises a FritzHttpInterfaceError if the device
        does not support the command or arguments. On missing rights a
        FritzAuthorizationError is raised.
        """
        return self.fh.fc.call_http(command, self.identifier, **kwargs)

    @property
    def identifier(self) -> str:
        return self.AIN  # type: ignore

    @property
    def is_han_fun_unit(self) -> bool:
        return self._bitmatch(HAN_FUN_UNIT_1) or self._bitmatch(HAN_FUN_UNIT_2)

    @property
    def is_bulb(self) -> bool:
        return self._bitmatch(LIGHT_BULB)

    @property
    def is_alarm_sensor(self) -> bool:
        return self._bitmatch(ALARM_SENSOR)

    @property
    def is_avm_button(self) -> bool:
        return self._bitmatch(AVM_BUTTON)

    @property
    def is_radiator_control(self) -> bool:
        return self._bitmatch(RADIATOR_CONTROL)

    @property
    def is_energy_sensor(self) -> bool:
        return self._bitmatch(ENERGY_SENSOR)

    @property
    def is_temperature_sensor(self) -> bool:
        return self._bitmatch(TEMPERATURE_SENSOR)

    @property
    def is_pluggable(self) -> bool:
        return self._bitmatch(PLUGGABLE)

    @property
    def is_avm_dect_repeater(self) -> bool:
        return self._bitmatch(AVM_DECT_REPEATER)

    @property
    def is_microphone(self) -> bool:
        return self._bitmatch(MICROPHONE)

    @property
    def is_switchable(self) -> bool:
        return self._bitmatch(SWITCHABLE)

    @property
    def is_adjustable(self) -> bool:
        return self._bitmatch(ADJUSTABLE)

    @property
    def is_color_bulb(self) -> bool:
        return self._bitmatch(COLOR_BULB)

    @property
    def is_blind(self) -> bool:
        return self._bitmatch(BLIND)

    @property
    def is_humidity_sensor(self) -> bool:
        return self._bitmatch(HUMIDITY_SENSOR)

    def update_device_information(self) -> None:
        """
        Triggers the router to update the device dependent
        instance-attributes.
        """
        self._extraxt_device_information_as_attributes(
            self.fh.get_device_information_by_identifier(self.identifier)
        )

    def get_basic_device_stats(self) -> dict:
        """
        Returns a dictionary of device statistics. The content depends on
        the actors supported by a device. The keys can be:

        key:            on actor:
        temperature     temperature sensor
        humidity        humidity sensor
        voltage         energy sensor
        power           energy sensor
        energy          energy sensor

        The corresponding values are also dictionaries having a common
        structure:

        key:        value:
        count       int: number of data points
        grid        int: time resolution in seconds
        datatime    datetime: timestamp of last data update
        data        list of data points (integers)

        The application can inspect the keys or use the properties like
        `is_energy_sensor` to check for actors provided by the device.
        If the usecase is just about the actors and not the data, the
        properties are much(!) faster.
        """
        response = self.call_http("getbasicdevicestats")
        return self.extract_basicdevicestats_response(response)

    @staticmethod
    def extract_basicdevicestats_response(response: dict) -> dict:
        """
        Converts the xml `response` and returns a dictionary with a
        datastructure described in the method `get_basic_device_stats()`
        """
        # implemented separately for testing.
        # 'stats' and 'datatime' are defined in the AVM xml-protocol
        # some types are dynamic therefore some 'type: ignore'.
        elements = {}
        content = response['content']
        root = etree.fromstring(content)
        for element in root:
            content = {}
            stats: etree.Element | None = element.find("stats")
            if stats is None:
                continue
            for key, value in stats.attrib.items():
                value = int(value)  # type: ignore
                if key == "datatime":
                    value = datetime.datetime.fromtimestamp(value)  # type: ignore
                content[key] = value
            # convert the csv-list of returned values from text to int.
            # on missing data dashes (-) may get returned.
            # this get catched and missing data are represented as `None`.
            content["data"] = []
            for item in stats.text.split(","):  # type: ignore
                try:
                    value = int(item)  # type: ignore
                except ValueError:
                    value = None  # type: ignore
                content["data"].append(value)
            elements[element.tag] = content
        return elements

    def get_switch_state(self) -> bool:
        """
        Returns a boolean indicating whether a switchable device is set
        to on (True) or off (False).
        """
        self.update_device_information()
        return self.SwitchState.lower() == 'on'  # type: ignore

    def set_switch(self, on: bool = True) -> None:
        """
        Set a switchable device to 'on' (True) or 'off' (False).
        """
        self.fh.set_switch(self.identifier, on)




"""
Module to read status-information from an AVM FritzBox.
"""

from __future__ import annotations

import datetime
from collections import namedtuple

from fritzconnection.core.processor import (
    Storage,
    InstanceAttributeFactory,
    processor,
    process_node,
)
from fritzconnection.core.utils import get_xml_root

from .fritzbase import AbstractLibraryBase
from .fritztools import (
    ArgumentNamespace,
    format_dB,
    format_num,
    format_rate,
)

DefaultConnectionService = namedtuple(
    "DefaultConnectionService", "prefix connection_service postfix"
)


def _integer_or_original(value):
    """
    Tries to convert value to an integer. Returns this integer on
    success, otherwise returns the original value.
    """
    try:
        return int(value)
    except ValueError:
        return value


@processor
class Event:
    """
    Represents an AVM DeviceLog entry with the subnodes given as
    class-attributes. The default types are strings as extracted from
    the xml-source.
    """

    id = None
    group = None
    date = None
    time = None
    msg = None

    @property
    def datetime(self):
        return datetime.datetime.strptime(f"{self.date}{self.time}", "%d.%m.%y%H:%M:%S")


class DeviceLog(Storage):
    """
    The AVM DeviceLog is a list of Event-nodes stored in the `events`
    instance attribute. But instances of DeviceLog are also iterables
    for the events.
    """

    Event = InstanceAttributeFactory(Event)

    def __init__(self, root):
        self.events = list()
        super().__init__(self.events)
        process_node(self, root)

    def __iter__(self):
        for event in self.events:
            yield event


class FritzStatus(AbstractLibraryBase):
    """
    Class for requesting status-information: up, down, ip, activity
    (bytes per second send/received). All parameters are optional. If
    given, they have the following meaning: `fc` is an instance of
    FritzConnection, `address` the ip of the Fritz!Box, `port` the port
    to connect to, `user` the username, `password` the password,
    `timeout` a timeout as floating point number in seconds, `use_tls` a
    boolean indicating to use TLS (default False).
    """

    @property
    def is_linked(self) -> bool:
        """
        A boolean whether the FritzBox is physically linked to
        the provider.
        """
        status = self.fc.call_action("WANCommonIFC", "GetCommonLinkProperties")
        return status["NewPhysicalLinkStatus"] == "Up"

    @property
    def is_connected(self) -> bool:
        """
        A boolean whether the FritzBox has established an
        internet-connection.
        """
        status = self.fc.call_action("WANIPConn", "GetStatusInfo")
        return status["NewConnectionStatus"] == "Connected"

    @property
    def external_ip(self) -> str:
        """The external v4 ip-address."""
        return self.fc.call_action("WANIPConn", "GetExternalIPAddress")[
            "NewExternalIPAddress"
        ]

    @property
    def external_ipv6(self) -> str:
        """The external v6 ip-address."""
        return self.external_ipv6_info["NewExternalIPv6Address"]

    @property
    def external_ipv6_info(self) -> dict:
        """
        Returns the ipv6 external address information as a dictionary with the keys:
        NewExternalIPv6Address                   out ->     string
        NewPrefixLength                          out ->     ui1
        NewValidLifetime                         out ->     ui4
        NewPreferedLifetime                      out ->     ui4
        """
        return self.fc.call_action("WANIPConn", "X_AVM_DE_GetExternalIPv6Address")

    @property
    def ipv6_prefix(self):
        """The internal v6 prefix."""
        return self.ipv6_prefix_info["NewIPv6Prefix"]

    @property
    def ipv6_prefix_info(self) -> dict:
        """
        Returns the ipv6 prefix information as a dictionary with the keys:
        NewIPv6Prefix                            out ->     string
        NewPrefixLength                          out ->     ui1
        NewValidLifetime                         out ->     ui4
        NewPreferedLifetime                      out ->     ui4
        """
        return self.fc.call_action("WANIPConn", "X_AVM_DE_GetIPv6Prefix")

    @property
    def connection_uptime(self) -> int:
        """Connection uptime in seconds."""
        status = self.fc.call_action("WANIPConn", "GetStatusInfo")
        return status["NewUptime"]

    @property
    def device_uptime(self) -> int:
        """Device uptime in seconds."""
        status = self.fc.call_action("DeviceInfo1", "GetInfo")
        return status["NewUpTime"]

    @property
    def str_uptime(self) -> str:
        """Connection uptime in human-readable format."""
        mins, secs = divmod(self.connection_uptime, 60)
        hours, mins = divmod(mins, 60)
        return "%02d:%02d:%02d" % (hours, mins, secs)

    @property
    def bytes_sent(self) -> int | str:
        """
        Total number of sent bytes. Returns an integer or, in case of
        failure, the original value which would be a string.
        """
        try:
            status = self.fc.call_action("WANCommonIFC1", "GetAddonInfos")
            value = status["NewX_AVM_DE_TotalBytesSent64"]
        except KeyError:
            # fallback for older FritzOS Versions not providing a 64 bit value:
            status = self.fc.call_action("WANCommonIFC1", "GetTotalBytesSent")
            value = status["NewTotalBytesSent"]
        return _integer_or_original(value)

    @property
    def bytes_received(self) -> int | str:
        """
        Total number of received bytes. Returns an integer or, in case of
        failure, the original value which would be a string.
        """
        try:
            status = self.fc.call_action("WANCommonIFC1", "GetAddonInfos")
            value = status["NewX_AVM_DE_TotalBytesReceived64"]
        except KeyError:
            # fallback for older FritzOS Versions not providing a 64 bit value:
            status = self.fc.call_action("WANCommonIFC1", "GetTotalBytesReceived")
            value = status["NewTotalBytesReceived"]
        return _integer_or_original(value)

    @property
    def transmission_rate(self) -> tuple[int, int]:
        """
        The upstream and downstream values as a tuple in bytes per
        second. Use this for periodical calling.
        """
        status = self.fc.call_action("WANCommonIFC1", "GetAddonInfos")
        upstream = status["NewByteSendRate"]
        downstream = status["NewByteReceiveRate"]
        return upstream, downstream

    @property
    def str_transmission_rate(self) -> tuple[str, str]:
        """
        Tuple of human-readable transmission rate in bytes. First item
        is upstream, second item downstream.
        """
        upstream, downstream = self.transmission_rate
        return format_num(upstream), format_num(downstream)

    @property
    def max_linked_bit_rate(self) -> tuple[int, int]:
        """
        Tuple with the maximum upstream- and downstream-rate
        of the physical link. The rate is given in bits/sec.
        """
        return self._get_max_bit_rate("WANCommonInterfaceConfig")

    @property
    def max_bit_rate(self) -> tuple[int, int]:
        """
        Tuple with the maximum upstream- and downstream-rate
        of the given connection. The rate is given in bits/sec.
        """
        return self._get_max_bit_rate("WANCommonIFC")

    def _get_max_bit_rate(self, servicename: str) -> tuple[int, int]:
        """
        internal method to get the upstream and downstream-rates for
        different services of the WANCommonInterfaceConfig1 ServiceType.
        """
        status = self.fc.call_action(servicename, "GetCommonLinkProperties")
        downstream = status["NewLayer1DownstreamMaxBitRate"]
        upstream = status["NewLayer1UpstreamMaxBitRate"]
        return upstream, downstream

    @property
    def max_byte_rate(self) -> tuple[float, float]:
        """
        Same as max_bit_rate but rate is given in bytes/sec.
        """
        upstream, downstream = self.max_bit_rate
        return upstream / 8.0, downstream / 8.0

    @property
    def str_max_linked_bit_rate(self) -> tuple[str, str]:
        """
        Human-readable maximum of the physical upstream- and
        downstream-rate in bits/sec. Value is a tuple, first item is
        upstream, second item is downstream.
        """
        upstream, downstream = self.max_linked_bit_rate
        return (
            format_rate(upstream, unit="bits"),
            format_rate(downstream, unit="bits"),
        )

    @property
    def str_max_bit_rate(self) -> tuple[str, str]:
        """
        Human-readable maximum of the upstream- and downstream-rate in
        bits/sec, as given by the provider. Value is a tuple, first item
        is upstream, second item is downstream.
        """
        upstream, downstream = self.max_bit_rate
        return (
            format_rate(upstream, unit="bits"),
            format_rate(downstream, unit="bits"),
        )

    def get_monitor_data(self, sync_group_index=0) -> dict:
        """
        Returns a dictionary with realtime data about the current up-
        and downstream rates.
        """
        monitor_data = self.fc.call_action(
            "WANCommonInterfaceConfig1",
            "X_AVM-DE_GetOnlineMonitor",
            NewSyncGroupIndex=sync_group_index,
        )
        for key, value in monitor_data.items():
            if isinstance(value, str) and "," in value:
                try:
                    items = [int(v) for v in value.split(",")]
                except (AttributeError, ValueError):
                    # ignore and keep value as is:
                    pass
                else:
                    monitor_data[key] = items  # type: ignore
        return monitor_data

    def reconnect(self) -> None:
        """Makes a reconnection with a new external ip."""
        self.fc.reconnect()

    @property
    def noise_margin(self) -> tuple[int, int]:
        """
        Tuple of noise margin. First item
        is upstream, second item downstream.
        """
        status = self.fc.call_action("WANDSLInterfaceConfig1", "GetInfo")
        upstream = status["NewUpstreamNoiseMargin"]
        downstream = status["NewDownstreamNoiseMargin"]
        return upstream, downstream

    @property
    def str_noise_margin(self) -> tuple[str, str]:
        """
        Human-readable noise margin in dB. Value is a tuple, first item
        is upstream, second item downstream.
        """
        upstream, downstream = self.noise_margin
        return format_dB(upstream), format_dB(downstream)

    @property
    def attenuation(self) -> tuple[int, int]:
        """
        Tuple of attenuation. First item
        is upstream, second item downstream.
        """
        status = self.fc.call_action("WANDSLInterfaceConfig1", "GetInfo")
        upstream = status["NewUpstreamAttenuation"]
        downstream = status["NewDownstreamAttenuation"]
        return upstream, downstream

    @property
    def str_attenuation(self) -> tuple[str, str]:
        """
        Human-readable attenuation in dB. Value is a tuple, first item
        is upstream, second item downstream.
        """
        upstream, downstream = self.attenuation
        return format_dB(upstream), format_dB(downstream)

    @property
    def upnp_enabled(self) -> bool:
        """
        Returns a boolean whether upnp is enabled or raises a
        FritzServiceError in case the service is not available.
        """
        status = self.fc.call_action("X_AVM-DE_UPnP1", "GetInfo")
        return status["NewEnable"]

    @property
    def device_has_mesh_support(self) -> bool:
        """
        True if the device supports mesh, otherwise False.
        """
        # check for the corresponding action
        # whether mesh is supported
        try:
            return "X_AVM-DE_GetMeshListPath" in self.fc.services["Hosts1"].actions
        except KeyError:
            # can happen if "Hosts1" is not known
            return False

    def get_device_info(self) -> ArgumentNamespace:
        """
        Returns an ArgumentNamespace with the attributes:

        manufacturer_name, manufacturer_oui, model_name, description,
        product_class, serial_number, software_version, hardware_version,
        spec_version, provisioning_code, up_time, device_log

        .. versionadded:: 1.10

        """
        return ArgumentNamespace(self.fc.call_action("DeviceInfo1", "GetInfo"))

    def get_avm_device_log(self, filter: str | None = None) -> DeviceLog:
        """
        The avm device log is a list of events with the attributes `id`,
        `group`, `date`, `time` and `msg` holding information like "DSL
        synchronization starting (training)" and other system messages.
        The Method returns a DeviceLog instance holding a list of Event
        instances. The DeviceLog instance is an iterable and can be
        used on a FritzStatus instance like:

        >>> device_log = fritzstatus.get_avm_device_log()
        >>> for event in device_log:
        >>>     print(event.datetime, event.msg)

        The returned events can be filtered by groups like 'sys', 'net',
        'fon', 'wlan' or 'usb'. To filter by a group provide the
        group-name as filter-argument.
        """
        result = self.fc.call_action("DeviceInfo1", "X_AVM-DE_GetDeviceLogPath")
        path = result["NewDeviceLogPath"]
        if filter:
            path = f"{path}&filter={filter}"
        url = f"{self.fc.address}:{self.fc.port}{path}"
        root_node = get_xml_root(url, session=self.fc.session)
        return DeviceLog(root_node)

    def get_cpu_temperatures(self) -> list[int]:
        """
        Returns a list of the last measured cpu-temperatures (Celsius).
        The most recent entry is the first one in the list.
        NOTE: this function call is experimental as it is based on a
        non-public API.
        """
        url = f"{self.fc.http_interface.router_url}/query.lua"
        payload = {"CPUTEMP": "cpu:status/StatTemperature"}
        response = self.fc.http_interface.call_url(url, payload)
        return list(map(int, response.json()["CPUTEMP"].split(",")))

    def get_default_connection_service(self) -> DefaultConnectionService:
        """
        Returns a namedtuple of type DefaultConnectionService:
        `prefix` -> str
        `device_connection` -> str (like "WANPPPConnection")
        `postfix` -> str
        """
        result = self.fc.call_action("Layer3Forwarding1", "GetDefaultConnectionService")
        prefix, connection_service, postfix = result[
            "NewDefaultConnectionService"
        ].split(".", 2)
        return DefaultConnectionService(prefix, connection_service, postfix)

    @property
    def connection_service(self) -> str:
        """
        The extracted connection_service from
        get_default_connection_service().
        """
        result = self.get_default_connection_service()
        return result.connection_service

    @property
    def update_available(self) -> str:
        """
        The new version number (as a string) if an update is available or an
        empty string if no update is avilable.
        """
        return self.fc.call_action("UserInterface1", "GetInfo")["NewX_AVM-DE_Version"]

    @property
    def has_wan_enabled(self) -> bool:
        """
        True if wan is enabled otherwise False.
        """
        return self.fc.call_action(self.connection_service, "GetInfo")["NewEnable"]

    @property
    def has_wan_support(self) -> bool:
        """
        True if the device supports a WAN interface.
        False otherwise.
        """
        return "Layer3Forwarding1" in self.fc.services




"""
Some helpers for the library.
"""

import math
import re
from types import SimpleNamespace


RE_UPPER_CASE = re.compile(r"([A-Z]+)")


def byte_formatter(value):
    """
    Gets a large integer als value and returns a tuple with the value as
    float and the matching dimension as string, i.e.
    >>> byte_formatter(242981246)
    (242.981246, 'MB')
    Expects positive integer as input. Negative numbers are interpreted
    as positive numbers. values < 1 are interpreted as 0.
    """
    dim = ['B', 'KB', 'MB', 'GB', 'TB']
    value = abs(value)
    if value < 1:
        log = 0
        num = 0
    else:
        log = min(int(math.log10(value) / 3), len(dim))
        num = value / 1000 ** log
    try:
        dimension = dim[log]
    except IndexError:
        dimension = 'PB'
    return num, dimension


def format_num(num, unit='bytes'):
    """
    Returns a human-readable string of a byte-value.
    If 'num' is bits, set unit='bits'.
    """
    num, dim = byte_formatter(num)
    if unit != 'bytes':
        dim += 'it'  # then its Bit by default
    return f'{num:3.1f} {dim}'


def format_rate(num, unit='bytes'):
    """
    Returns a human-readable string of a byte/bits per second.
    If 'num' is bits, set unit='bits'.
    """
    return format_num(num, unit=unit) + '/s'


def format_dB(num):
    """
    Returns a human-readable string of dB. The value is divided
    by 10 to get first decimal digit
    """
    num /= 10
    return f'{num:3.1f} {"dB"}'


class ArgumentNamespace(SimpleNamespace):
    """
    Namespace object that also behaves like a dictionary, but is not
    iterable.

    Usecase is as a wrapper for the dictionary returned from
    `FritzConnection.call_action()`. This dictionary has keys named
    "arguments" as described by the AVM documentation, combined with
    values as the corresponding return values. Instances of
    `ArgumentNamespace` can get used to extract a subset of this
    dictionary and transfer the Argument-names to more readable
    ones. For example consider that `fc` is a FritzConnection instance.
    Then the following call: ::

        >>> result = fc.call_action("DeviceInfo1", "GetInfo")

    will return a dictionary like: ::

        {'NewManufacturerName': 'AVM',
         'NewManufacturerOUI': '00040E',
         'NewModelName': 'FRITZ!Box 7590',
         'NewDescription': 'FRITZ!Box 7590 154.07.29',
         'NewProductClass': 'AVMFB7590',
         'NewSerialNumber': '989BCB2xxxxx',
         'NewSoftwareVersion': '154.07.29',
         'NewHardwareVersion': 'FRITZ!Box 7590',
         'NewSpecVersion': '1.0',
         'NewProvisioningCode': '000.044.004.000',
         'NewUpTime': 9516949,
         'NewDeviceLog': 'long string here ...'}

    In case that just the model name and serial number are of interest,
    and should have better names, first define a mapping: ::

        mapping = {
            "modelname": "NewModelName",
            "serialnumber": "NewSerialNumber"
        }

    and use this `mapping` with the dictionary `result` to create an
    `ArgumentNamespace` instance: ::

        >>> info = ArgumentNamespace(result, mapping)

    The `info` instance can now get used like a namespace object and
    like a dictionary: ::

        >>> info.serialnumber
        '989BCB2xxxxx'

        >>> info['modelname']
        'FRITZ!Box 7590'

    If no mapping is given, then `ArgumentNamespace` will consume the
    provided dictionary converting all keys from "MixedCase" style to
    "snake_case" (PEP 8) and removes the leading "new" originating from
    the AVM "New" prefix for all keys: ::

        >>> info = ArgumentNamespace(result)
        >>> info.up_time
        9516949

    Setting the flag `suppress_new` to `False` will keep the prefix: ::

        >>> info = ArgumentNamespace(result, suppress_new=False)
        >>> info.new_up_time
        9516949

    To just extract a subset of `result` provide a sequence of key-names
    for the `extract` argument: ::

        >>> extract = "NewSerialNumber", "NewModelName"
        >>> info = ArgumentNamespace(result, extract=extract)
        >>> info.serial_number
        '989BCB2xxxxx'
        >>> info.model_name
        'FRITZ!Box 7590'

    If both arguments `mapping` and `extract` are given, `mapping` has
    precedence and `extract` gets ignored.

    .. versionadded:: 1.10

    """

    def __init__(self, source, mapping=None, extract=None, suppress_new=True):
        if mapping is None:
            keys = extract if extract else source.keys()
            mapping = {
                self.rewrite_argument(key, suppress_new): key for key in keys
            }
        super().__init__(
            **{name: source[attribute] for name, attribute in mapping.items()}
        )

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __len__(self):
        # should provide len() as dict-like object
        return len(self.__dict__)

    @staticmethod
    def rewrite_argument(name: str, suppress_new: bool = True) -> str:
        """
        Rewrite `name` from MixedCase to snake_case (PEP 8). So i.e.
        "MixedCase" gets converted to "mixed_case". AVM standard
        argument names starting with "New" like "NewMixedCase" will get
        converted to "new_mixed_case". If `suppress_new` is `True` (the
        default) the "new"-prefix will get removed, so "NewMixedCased"
        will get converted to "mixed_case". Consecutive upper-case
        characters are handled as a group: "ManufacturerOUI" ->
        "manufacturer_oui".

        .. versionadded:: 1.10

        """
        new = "new_"
        result = RE_UPPER_CASE.sub(r"_\1", name).lower()
        if result.startswith("_"):
            result = result[1:]
        if suppress_new and result.startswith(new):
            result = result[len(new):]
        return result



"""
Module defining fritzconnection specific exceptions.
"""

# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer
#
#
# Exception Inheritance:
# ----------------------
#
# FritzConnectionException
#                 |
#                 |--> ActionError --> FritzActionError
#                 |--> ServiceError --> FritzServiceError
#                 |
#                 |--> FritzAuthorizationError
#                 |--> FritzHttpInterfaceError
#                 |
#                 |--> FritzResourceError
#                 |
#                 |--> FritzArgumentError
#                 |       |
#                 |       |--> FritzArgumentValueError
#                 |               |
#                 |               |--> FritzArgumentStringToShortError
#                 |               |--> FritzArgumentStringToLongError
#                 |               |--> FritzArgumentCharacterError
#                 |
#                 |--> FritzInternalError
#                 |       |
#                 |       |--> FritzActionFailedError
#                 |       |--> FritzOutOfMemoryError
#                 |
#                 |--> FritzSecurityError
#                 |
#                 |-->|--> FritzLookUpError
#                 |   |
# KeyError -------+-->|
#                 |
#                 |
#                 |-->|--> FritzArrayIndexError
#                     |
# IndexError -------->|
#
#


# export all Exceptions on * imports
__all__ = [
    'FritzConnectionException',
    'FritzActionError',
    'FritzActionFailedError',
    'FritzHttpInterfaceError',
    'FritzArgumentCharacterError',
    'FritzArgumentError',
    'FritzArgumentStringToLongError',
    'FritzArgumentStringToShortError',
    'FritzArgumentValueError',
    'FritzArrayIndexError',
    'FritzAuthorizationError',
    'FritzInternalError',
    'FritzLookUpError',
    'FritzOutOfMemoryError',
    'FritzResourceError',
    'FritzSecurityError',
    'FritzServiceError',
]


class FritzConnectionException(Exception):
    """Base Exception for communication errors with the Fritz!Box"""


class ActionError(FritzConnectionException):
    """
    Exception raised by calling non-existing actions.
    Legacy Exception. Use FritzActionError instead.
    """


class ServiceError(FritzConnectionException):
    """
    Exception raised by calling non-existing services.
    Legacy Exception. Use FritzServiceError instead.
    """


class FritzServiceError(ServiceError):
    """Exception raised by calling non-existing services."""


class FritzActionError(ActionError):
    """Exception raised by calling non-existing actions."""


class FritzHttpInterfaceError(FritzConnectionException):
    """
    Exception raised on calling the aha-interface and getting an
    response with a status-code other than 200.
    """


class FritzResourceError(FritzConnectionException):
    """
    Exception raised if a description file like an igddesc-file is not
    provided by the router. This Exception is used internally and
    logged, but not reraised for the user of the library.
    """


class FritzArgumentError(FritzConnectionException):
    """Exception raised by invalid arguments."""


class FritzArgumentValueError(FritzArgumentError):
    """
    Exception raised by arguments with invalid values.
    Inherits from the more generic FritzArgumentError.
    """


class FritzArgumentStringToShortError(FritzArgumentValueError):
    """
    Exception raised by arguments with invalid string length for the
    string being to short. Inherits from the more generic
    FritzArgumentValueError.
    """


class FritzArgumentStringToLongError(FritzArgumentValueError):
    """
    Exception raised by arguments with invalid string length for the
    string being to long. Inherits from the more generic
    FritzArgumentValueError.
    """


class FritzArgumentCharacterError(FritzArgumentValueError):
    """
    Exception raised by arguments with invalid characters.
    Inherits from the more generic FritzArgumentValueError.
    """


class FritzInternalError(FritzConnectionException):
    """Exception raised by panic in the box."""


class FritzActionFailedError(FritzInternalError):
    """
    Exception raised by the box unable to execute the action properly.
    Inherits from the more generic FritzInternalError.
    """


class FritzOutOfMemoryError(FritzInternalError):
    """
    Exception raised by memory shortage of the box.
    Inherits from the more generic FritzInternalError.
    """


class FritzSecurityError(FritzConnectionException):
    """Authorization error or wrong security context."""


class FritzLookUpError(FritzConnectionException, KeyError):
    """
    Lookup for id or entry in existing internal array failed.
    Inherits from KeyError.
    So KeyError can also be used for exception handling.
    """


class FritzArrayIndexError(FritzConnectionException, IndexError):
    """
    Addressing an entry in an internal array by index failed.
    Inherits from IndexError.
    So IndexError can also be used for exception handling.
    """

class FritzAuthorizationError(FritzConnectionException):
    """
    Authentication error.
    Not allowed to access the box at all.
    """

# Collection of error codes and corresponding exceptions:
# (the error codes are defined by AVM)
FRITZ_ERRORS = {
    '401': FritzActionError,
    '402': FritzArgumentError,
    '501': FritzActionFailedError,
    '600': FritzArgumentValueError,
    '603': FritzOutOfMemoryError,
    '606': FritzSecurityError,
    '713': FritzArrayIndexError,
    '714': FritzLookUpError,
    '801': FritzArgumentStringToShortError,
    '802': FritzArgumentStringToLongError,
    '803': FritzArgumentCharacterError,
    '820': FritzInternalError,
}



"""
fritzhttp.py

Access the AVM Fritz!Box AHA-HTTP-Interface
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import hashlib
from http import HTTPStatus
from http.client import HTTP_PORT
from xml.etree import ElementTree as etree

from fritzconnection.core.exceptions import (
    FritzHttpInterfaceError,
    FritzAuthorizationError,
)


URL_LOGIN = "/login_sid.lua?version=2"
URL_HOMEAUTOSWITCH = "/webservices/homeautoswitch.lua"
PBKDF2_CHALLENGE_INDICATOR = "2$"


class FritzHttp:
    """
    Implementation for the AVM AHA-HTTP-Inferface.

    The current implementation does not handle a blocktime timeout so
    far. This is because the communication is based on a
    fritzconnection-session and the proper credentials are already
    handled there.

    There may be the side-effect that someone else messes up with the
    login of the human web-interface while this script is running. In
    this case the aha-interface login will not return a valid sid until
    blocktime runs out.
    """
    def __init__(self, fc):
        self.fc = fc  # the active fritzconnection instance
        self.sid = None

    @property
    def remote_port(self):
        """
        Provides the configurable https port for the aha-interface as int.
        """
        if self.fc.address.startswith("https"):
            data = self.fc.call_action("X_AVM-DE_RemoteAccess1", "GetInfo")
            return int(data["NewPort"])  # provide same type as HTTP_PORT
        return HTTP_PORT

    @property
    def router_url(self):
        """Returns the combination of router address and port."""
        return f"{self.fc.address}:{self.remote_port}"

    @property
    def login_url(self):
        """The login-url including protocol and configurable port."""
        return f"{self.router_url}{URL_LOGIN}"

    @property
    def homeauto_url(self):
        """The homeauto-url including protocol and configurable port."""
        return f"{self.router_url}{URL_HOMEAUTOSWITCH}"

    def execute(self, command=None, identifier=None, **kwargs):
        """
        Send the command and the optional identifier to the
        http-interface and returns a tuple with the content-type and the
        response-text as is. On error raises a FritzAuthorizationError
        if the error code was 403 otherwise raises a generic
        FritzConnectionException with the corresponding error-code.

        The `command` is a string like 'getswitchlist' or
        'getbasicdevicestats' according to the AVM AHA documentation.

        The `identifier` is a string, representing a device-ain.
        """
        payload = {"switchcmd": command, "ain": identifier}
        payload.update(kwargs)
        response = self.call_url(self.homeauto_url, payload)
        return response.headers.get('content-type'), response.text

    def call_url(self, url, payload):
        """
        Makes a call to the router with the provided url. Returns the
        request object in case of success. Otherwise a
        FritzHttpInterfaceError will get raised.

        Beside the public API documented by AVM this method allows calls
        to undocumented APIs serving the router web-interface or
        providing other data.

        WARNING: For a reliable application it is highly discouraged to
        use undocumented endpoints because they can change any time without
        notice. So an application may not survive a router OS update.
        """
        for sid in self._get_sid():
            payload['sid'] = sid
            with self.fc.session.get(url, params=payload) as response:
                if response.status_code == HTTPStatus.OK:
                    return response

        msg = f"Request failed: http error code '{response.status_code}'"
        if response.status_code == HTTPStatus.FORBIDDEN:
            # can happen if FritzConnection was initialized
            # without a password.
            raise FritzAuthorizationError(msg)
        # This can be from the 400 or 500 error-family.
        # Most often these errors are triggered by a malformed payload,
        # therefore include the payload in the message:
        msg = f"{msg}, payload: {payload}"
        raise FritzHttpInterfaceError(msg)

    def _get_sid(self):
        """
        Generator to provide the sid two times in case the first try
        failed. This can happen on an invalide or expired sid. In this
        case the sid gets regenerated for the second try.
        """
        if self.sid is None:
            # a session id of None can lead to irritation
            self._set_sid_from_box()
        yield self.sid
        self._set_sid_from_box()
        yield self.sid

    def _set_sid_from_box(self):
        """
        Read a session id from the box and store it in self.sid
        As long as self.sid holds a valid sid, the user is logged in.
        """
        with self.fc.session.get(self.login_url) as response:
            challenge = etree.fromstring(response.text).find('Challenge').text
        if challenge.startswith(PBKDF2_CHALLENGE_INDICATOR):
            challenge_hash = self._get_pbkdf2_hash(challenge)
        else:
            challenge_hash = self._get_md5_hash(challenge)
        self.sid = self._request_sid(challenge_hash)

    def _get_pbkdf2_hash(self, challenge):
        """Returns the vendor-recommended pbkdf2 challenge hash."""
        _, iterations_1, salt_1, iterations_2, salt_2 = challenge.split('$')
        static_hash = hashlib.pbkdf2_hmac(
            "sha256",
            self.fc.soaper.password.encode(),
            bytes.fromhex(salt_1),
            int(iterations_1)
        )
        dynamic_hash = hashlib.pbkdf2_hmac(
            "sha256",
            static_hash,
            bytes.fromhex(salt_2),
            int(iterations_2)
        )
        return f"{salt_2}${dynamic_hash.hex()}"

    def _get_md5_hash(self, challenge):
        """Returns the legathy md5 challenge hash."""
        hash = hashlib.md5(
            f"{challenge}-{self.fc.soaper.password}".encode("utf-16-le")
        )
        return f"{challenge}-{hash.hexdigest()}"

    def _request_sid(self, challenge_hash):
        """
        Takes the challenge_hash to request and return a new session id.
        """
        # TODO: handle blocktime
        with self.fc.session.post(
            self.login_url,
            data={"username": self.fc.soaper.user, "response": challenge_hash},
            headers={"Content-Type": "application/x-www-form-urlencoded"}
        ) as response:
            root = etree.fromstring(response.text)
            sid_node = root.find("SID")
            return sid_node.text



"""
Common functions for other core-modules.
"""

import os
import re
from xml.etree import ElementTree as etree

import requests

from .exceptions import FritzConnectionException, FritzResourceError
from .logger import fritzlogger


NS_REGEX = re.compile("({(?P<namespace>.*)})?(?P<localname>.*)")
VALUES_TRUE = {"true", "on", "1"}
VALUES_FALSE = {"false", "off", "0"}


def localname(node):
    if callable(node.tag):
        return "comment"
    m = NS_REGEX.match(node.tag)
    return m.group('localname')


def get_content_from(url, timeout=None, session=None):
    """
    Returns text from a get-request for the given url. In case of a
    secure request (using TLS) the parameter verify is set to False, in order to
    disable certificate verifications. As the Fritz!Box creates a
    self-signed certificate for use in the LAN, encryption will work but
    verification will fail.
    """
    def handle_response(response):
        fritzlogger.debug(response.text)
        ct = response.headers.get("Content-type")
        if ct == "text/html":
            message = f"Unable to retrieve resource '{url}' from the device."
            # this error will get catched, because it may happen depending
            # on the used router model, without doing any harm.
            # However it's logged on INFO level:
            fritzlogger.info(message)
            raise FritzResourceError(message)
        return response.text

    def do_request():
        fritzlogger.debug(f"requesting: {url}")
        if session:
            with session.get(url, timeout=timeout) as response:
                return handle_response(response)
        response = requests.get(url, timeout=timeout, verify=False)
        return handle_response(response)

    try:
        return do_request()
    except requests.exceptions.ConnectionError as err:
        message = f"Unable to get a connection: {err}"
        # that's an error worth logging:
        fritzlogger.error(message)
        # raise from None because the message holds the
        # proper information about the connection failure:
        raise FritzConnectionException(message) from None


def get_xml_root(source, timeout=None, session=None):
    """
    Function to help migrate from lxml to the standard-library xml-package.

    'source' must be a string and can be a xml-string, a uri or a file
    name. `timeout` is an optional parameter limiting the time waiting
    for a router response.
    In all cases this function returns a xml.etree.Element instance
    which is the root of the parsed tree.
    """
    if source.startswith("http://") or source.startswith("https://"):
        # it's an uri, use requests to get the content
        source = get_content_from(source, timeout=timeout, session=session)
    elif not source.startswith("<"):
        # assume it's a filename
        with open(source) as fobj:
            source = fobj.read()
    return etree.fromstring(source)


def boolean_from_string(value):
    """
    Takes a value as a string and converts it to a boolean or None. The
    string could be "true" or "false" in upper-, lower- and mixed-case.
    Also "on", "off" and "0", "1" are converted. If the value can not
    converted a ValueError gets raised. If the value does not support
    the .lower() method, an AttributeError gets raised.
    """
    lower_value = value.lower()
    if lower_value in VALUES_TRUE:
        return True
    if lower_value in VALUES_FALSE:
        return False
    raise ValueError(f"can't convert '{lower_value}' to a boolean.")


def get_boolean_from_string(value, default=None):
    """
    Same as `boolean_from_string` but returns the `default` argument
    instead of raising an exception.
    """
    try:
        return boolean_from_string(value)
    except (AttributeError, ValueError):
        return default


def get_bool_env(key, default=None):
    """
    Return the value of the environment variable key converted to a
    boolean if it exists, or default if it doesn’t or can't get
    converted to a boolean. keys convertable to a boolean are "true",
    "on", "1" and "false", "off", "0".
    """
    value = os.getenv(key)
    return get_boolean_from_string(value, default)



"""
fritzconnection

library to communicate with the AVM Fritz!Box
by the TR-064 protocoll and allowes access via the http-interface.
Provides also an interface for realtime call-monitoring.

for documentation refer:
https://avm.de/service/schnittstellen/
https://fritzconnection.readthedocs.io/
"""

# unused shortcut import are intended:
# ruff: noqa: F401

__version__ = "1.15.0"

# import shortcuts
from .core.fritzconnection import FritzConnection
from .core.fritzmonitor import FritzMonitor



"""
utils.py

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

import argparse
import os

from ..core.fritzconnection import (
    FritzConnection,
    FRITZ_CACHE_FORMAT_PICKLE,
    FRITZ_IP_ADDRESS,
    FRITZ_TCP_PORT,
    FRITZ_ENV_USERNAME,
    FRITZ_ENV_PASSWORD,
    FRITZ_ENV_USECACHE,
    FRITZ_ENV_CACHEDIRECTORY,
    FRITZ_ENV_CACHE_FORMAT,
)
from ..core.utils import get_bool_env
from .. import __version__


def print_header(instance):
    print(f'\nfritzconnection v{__version__}')
    if isinstance(instance, FritzConnection):
        print(instance)
    else:
        print(instance.fc)
    print()


def print_common_exception_message(error_object):
    print(error_object)
    print(
        "\nSeems you forgot to provide the user and/or the password."
        "\nYou can provide these with the flag -u and -p or store them"
        "\nin the environment as FRITZ_USERNAME and FRITZ_PASSWORD."
        "\n(Environment changes will get recognized by new processes.)\n"
    )


def get_instance(cls, args):
    # -y implies -x:
    if not args.verify_cache:
        args.use_cache = True
    return cls(
        address=args.address,
        port=args.port,
        user=args.username,
        password=args.password,
        use_tls=args.encrypt,
        use_cache=args.use_cache,
        verify_cache=args.verify_cache,
        cache_directory=args.cache_directory,
        cache_format=args.cache_format,
    )


def get_cli_arguments(scan_additional_arguments=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--ip-address',
                        nargs='?', default=FRITZ_IP_ADDRESS, const=None,
                        dest='address',
                        help='Specify ip-address of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_IP_ADDRESS)
    parser.add_argument('--port',
                        nargs='?', default=None, const=None,
                        help='Port of the FritzBox to connect to. '
                             'Default: %s' % FRITZ_TCP_PORT)
    parser.add_argument('-u', '--username',
                        nargs='?', default=os.getenv(FRITZ_ENV_USERNAME, None),
                        help='Fritzbox authentication username')
    parser.add_argument('-p', '--password',
                        nargs='?', default=os.getenv(FRITZ_ENV_PASSWORD, None),
                        help='Fritzbox authentication password')
    parser.add_argument('-e', '--encrypt',
                        nargs='?', default=False, const=True,
                        help='Flag: use secure connection (TLS)')
    parser.add_argument('-x', '--use-cache',
                        default=get_bool_env(FRITZ_ENV_USECACHE, False),
                        action="store_true",
                        dest='use_cache',
                        help='Flag: use api cache (e[x]cellerate: speed-up subsequent '
                             'instantiations)'
                        )
    parser.add_argument('-y', '--suppress-cache-verification',
                        action='store_false',
                        default=True,
                        dest='verify_cache',  # needs inverted boolean to suppress.
                        help='Flag: suppress cache verification, implies -x'
                        )
    parser.add_argument('--cache-format',
                        nargs='?', default=os.getenv(
                            FRITZ_ENV_CACHE_FORMAT, FRITZ_CACHE_FORMAT_PICKLE
                        ),
                        dest='cache_format',
                        help="cache-file format: json|pickle (default: pickle)"
                        )
    parser.add_argument('--cache-directory',
                        nargs='?',
                        default=os.getenv(FRITZ_ENV_CACHEDIRECTORY, None),
                        const=None,
                        dest='cache_directory',
                        help="path to cache directory (default: ~.fritzconnection)"
                        )
    if scan_additional_arguments:
        scan_additional_arguments(parser)
    args = parser.parse_args()
    return args



"""
Module to access to the Mesh Topology
See https://avm.de/service/schnittstellen/ "Mesh-Topologie"
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Authors: Klaus Bremer

# code partly to dynamic for a useful mypy run:
# mypy: disable-error-code="attr-defined"

from __future__ import annotations

from .fritzhosts import FritzHosts
from .fritzbase import AbstractLibraryBase


class Connection:
    """
    Represents a connection from one device to another device.
    This is basically a convenience wrapper for InterfaceLink
    exposing the devices and some instance attributes in a more
    comfortable way.

    self.source is the origin device
    self.target is the connected device
    self.type is the connection type (like "WLAN")
    self.state is the connection state (like "DISCONNECTED", "CONNECTED")

    self.max_rx, max_tx, cur_rx and cur_tx are the maximum and
    current throughput data rates in kbit/sec.

    tx is the upload-throughput from source to target, even if the
    source is represented by the interface-link node 2.
    """

    def __init__(self, source, target, interface_link):
        self.source = source
        self.target = target
        self.interface_link = interface_link

    @property
    def type(self) -> str | None:
        return self._get_interface_attribute("type")

    @property
    def state(self) -> str | None:
        return self._get_interface_attribute("state")

    @property
    def max_rx(self) -> int | None:
        return self._get_transfer_rate("max_data_rate_rx")

    @property
    def max_tx(self) -> int | None:
        return self._get_transfer_rate("max_data_rate_tx")

    @property
    def cur_rx(self) -> int | None:
        return self._get_transfer_rate("cur_data_rate_rx")

    @property
    def cur_tx(self) -> int | None:
        return self._get_transfer_rate("cur_data_rate_tx")

    def _get_transfer_rate(self, name):
        """
        Adapt the rate to the link direction: 'tx' is the transfer rate
        from node 1 to node 2 which is ok, if node 1 represents the
        'source' node. In case it is the other way around the commands
        'tx' and 'rx' must get exchanged.
        """
        if self.interface_link.source_index == 2:
            name, direction = name.rsplit("_", 1)
            direction = "rx" if direction == "tx" else "tx"
            name = f"{name}_{direction}"
        return self._get_interface_attribute(name)

    def _get_interface_attribute(self, name, default=None):
        return getattr(self.interface_link, name, default)


class InterfaceLink:
    """
    Represents a node-link from an Interface of a device
    to another Interface of another device.
    (These are the vertexes of the mesh.)

    Instance attributes are all elements described as "node_links.properties"
    in the AVM documentation and accessible by their names, i.e. like
    'self.last_connected' in case the UNIX timestamp is requested.
    """

    def __init__(self, data: dict, interface: Interface):
        self.__dict__.update(data)
        self.interface = interface
        # The starting point is assumed to be the device, which is
        # the owner of self.interface.
        if self.node_interface_1_uid == self.interface.uid:
            self.source_index = 1
            self.target_index = 2
        else:
            self.source_index = 2
            self.target_index = 1

    def __repr__(self) -> str:
        return f"InterfaceLink:  {self.uid:<8}target-interface:  {self.node_interface_2_uid}"

    def __str__(self) -> str:
        connection = Connection(self.source, self.target, self)
        if connection.cur_tx is not None:
            throughput = connection.cur_tx // 1000  # from kB/s to MB/s
        else:
            throughput = None  # not known
        return f"from {self.source.name} to {self.target.name} (-> {throughput} MBit/s)"

    @property
    def source(self) -> Device:
        return self._get_connected_device(target=False)

    @property
    def target(self) -> Device:
        return self._get_connected_device(target=True)

    def get_connection(self) -> Connection:
        """
        Returns a connection instance describing a connection
        from one device to another device.
        """
        return Connection(self.source, self.target, self)

    def _get_connected_device(self, target=True):
        """
        Returns the device connected by this link,
        either the target or source device.
        """
        index = self.target_index if target else self.source_index
        device_id = getattr(self, f"node_{index}_uid")
        return self.interface.device.mesh.get_device_by_id(device_id)


class Interface:
    """
    An interface represents a physical infrastructure of a device
    to connect with other devices.
    An interface is part of a device and able to connect to another
    interface of another device.
    A connection is represented by the InterfaceLink class.
    """

    def __init__(self, data: dict, device: Device):
        self.__dict__.update(data)
        self.interface_links = [
            InterfaceLink(link, self) for link in self.node_links
        ]
        self.device = device

    def __repr__(self) -> str:
        representation = (
            f"Interface:  mac={self.mac:<20}"
            f"uid={self.uid:<8}"
            f"type={self.type:<8}"
        )
        if self.name:
            representation = f"{representation}name={self.name}"
        return representation

    def __str__(self) -> str:
        interface = repr(self)
        if self.interface_links:
            space = " " * 8
            links = f"\n{space}".join(str(link) for link in self.interface_links)
            interface = f"{interface}\n{space}{links}"
        return interface

    @property
    def mac(self) -> str:
        """mac address of the interface"""
        return self.mac_address

    def get_connections(self):
        """
        Returns a list of Connection objects describing all devices
        connected by this interface.
        """
        return [link.get_connection() for link in self.interface_links]


class Device:
    """
    A Device represents a physical item in the mesh, like a router,
    repeater, a laptop, a cell phone and other devices.
    (This is a node of the mesh.)

    Instance attributes created dynamical according to the
    AVM mesh topology schema: https://avm.de/service/schnittstellen/

    For convenience there are some properties and methods to access
    the data in a more mnemonic way.
    """

    def __init__(self, data: dict, mesh: FritzMeshTopology):
        self.__dict__.update(data)
        self.interfaces = [
            Interface(interface_data, self)
            for interface_data in self.node_interfaces
        ]
        self.mesh = mesh

    def __repr__(self) -> str:
        return f"Device:  mac={self.mac:<20}uid={self.uid:<8}{self.name}"

    def __str__(self) -> str:
        representation = repr(self)
        space = " "  * 4
        interfaces = f"\n{space}".join(str(interface) for interface in self.interfaces)
        return f"{representation}\n{space}{interfaces}"

    @property
    def name(self) -> str:
        return self.device_name

    @property
    def model(self) -> str:
        return self.device_model

    @property
    def vendor(self) -> str:
        return self.device_manufacturer

    @property
    def mac(self) -> str:
        return self.device_mac_address

    def get_connections(self):
        """
        Returns a list of Connection objects describing all
        devices connected to this device.
        """
        connections = []
        for interface in self.interfaces:
            connections.extend(interface.get_connections())
        return connections


class FritzMeshTopology(AbstractLibraryBase):
    """
    Represents the mesh topology with nodes representing the devices.

    """

    def __init__(self, fc=None, *args, **kwargs):
        super().__init__(fc, *args, **kwargs)
        self._fritzhosts = FritzHosts(self.fc)
        self.topology = {}
        self.nodes = {}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}: "
            f"schema_version={self.schema_version}, "
            f"registered Devices: {self.number_of_devices}"
        )

    def __str__(self) -> str:
        nodes = "\n\n".join(str(device) for device in self.devices)
        return f"{repr(self)}\n{nodes}"

    @property
    def schema_version(self) -> str:
        return self.topology.get("schema_version", "unknown")

    @property
    def number_of_devices(self) -> int:
        return len(self.nodes)

    @property
    def devices(self) -> list[Device]:
        return list(self.nodes.values())

    def get_device_by_id(self, uid: str) -> Device:
        """
        Returns a device by id. The id is the node-uid like "n-1".
        The node-uids are fixed to devices after calling
        `load_topology()` but may change if `load_topology()` gets
        called again.
        """
        return self.nodes[uid]

    def load_topology(self):
        """
        Load the topology from the router.
        """
        self.topology = self._fritzhosts.get_mesh_topology(raw=False)
        self.nodes = {
            node["uid"]: Device(node, self)
            for node in self.topology.get("nodes", ())
        }



"""
fritzinspection.py

Module to inspect the FritzBox API for available services and actions.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

import datetime

from ..core.exceptions import FritzAuthorizationError
from ..core.fritzconnection import FritzConnection
from .utils import (
    get_cli_arguments,
    get_instance,
    print_header,
    print_common_exception_message
)


class FritzInspection:
    """
    Class for cli use to inspect available services and according
    actions of the given device (the Fritz!Box).
    """
    # pylint: disable=invalid-name  # self.fc is ok.

    def __init__(self, fc):
        self.fc = fc

    def view_servicenames(self):
        """Send all known service names to stdout."""
        print('Servicenames:')
        for service_name in self.fc.services:
            print('{:20}{}'.format('', service_name))

    def view_actionnames(self, service_name, view_arguments=False):
        """Send all action names of the given service to stdout."""
        print('\n{:<20}{}'.format('Servicename:', service_name))
        print('Actionnames:')
        try:
            service = self.fc.services[service_name]
        except KeyError as err:
            print(f'Error: Invalid Servicename {err}')
        else:
            for action_name in service.actions:
                print('{:20}{}'.format('', action_name))
                if view_arguments:
                    action = service.actions[action_name]
                    for argument in sorted(action.arguments.keys()):
                        print('{:24}- {}'.format('', argument))
                    print()

    def view_actionarguments(self, service_name, action_name):
        """Send all action names of the given service to stdout."""
        try:
            service = self.fc.services[service_name]
        except KeyError as err:
            print(f'Error: Invalid Servicename {err}')
            return
        try:
            action = service.actions[action_name]
        except KeyError as err:
            print(f'Error: Invalid Actionname {err}')
            return
        print('\n{:<20}{}'.format('Service:', service_name))
        print('{:<20}{}'.format('Action:', action_name))
        print('Parameters:\n')
        print('    {:38}{:14}{}\n'.format('Name', 'direction', 'data type'))
        for argument in action.arguments.values():
            if argument.direction == 'in':
                direction = '-> in'
            else:
                direction = '   out ->'
            var = service.state_variables.get(argument.relatedStateVariable, '')
            line = f'    {argument.name:38}{direction:14}{var.dataType}'
            print(line)

    def view_complete_api(self):
        """
        Send the complete api to stdout.

        This can be a lengthy output that may be redirected to a file.
        """
        print()
        system_info = self.fc.device_manager.system_info
        # some devices may return None as system_info
        if system_info:
            print(
                f"system : {system_info[-1]}\n"
                f"build  : {system_info[-2]}\n"
                f"hw-code: {system_info[0]}"
            )
        now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print(f"Report date: {now}")
        for service_name, service in self.fc.services.items():
            if service_name == 'any1':
                continue
            print()
            print('=' * 65)
            for action_name in service.actions:
                self.view_actionarguments(service_name, action_name)


def add_arguments(parser):
    parser.add_argument('-r', '--reconnect',
                        action='store_true',
                        help='Reconnect and get a new ip')
    parser.add_argument('-R', '--reboot',
                        action='store_true',
                        help='Reboot the router')
    parser.add_argument('-s', '--services',
                        action='store_true',
                        help='List all available services')
    parser.add_argument('-S', '--serviceactions',
                        nargs=1,
                        help='List actions for the given service: <service>')
    parser.add_argument('-a', '--servicearguments',
                        nargs=1,
                        help='List arguments for the actions of a '
                             'specified service: <service>.')
    parser.add_argument('-A', '--actionarguments',
                        nargs=2,
                        help='List arguments for the given action of a '
                             'specified service: <service> <action>. '
                             'Lists also direction and data type of the '
                             'arguments.')
    parser.add_argument('-c', '--complete',
                        nargs='?', default=False, const=True,
                        help='List the complete api of the router')


def run_inspector(inspector, args):
    print_header(inspector.fc)
    if args.services:
        inspector.view_servicenames()
    elif args.serviceactions:
        inspector.view_actionnames(args.serviceactions[0])
    elif args.servicearguments:
        inspector.view_actionnames(args.servicearguments[0], view_arguments=True)
    elif args.actionarguments:
        inspector.view_actionarguments(args.actionarguments[0],
                                       args.actionarguments[1])
    elif args.complete:
        inspector.view_complete_api()
    elif args.reconnect:
        inspector.fc.reconnect()
        print('reconnect the router.')
    elif args.reboot:
        inspector.fc.reboot()
        print('reboot the router.')
    print()


def execute():
    args = get_cli_arguments(add_arguments)
    fc = get_instance(FritzConnection, args)
    inspector = FritzInspection(fc=fc)
    run_inspector(inspector, args)


def main():
    try:
        execute()
    except FritzAuthorizationError as err:
        print_common_exception_message(err)


if __name__ == '__main__':
    main()



"""
Module handling the SOAP based communication with the router.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import datetime
from logging import DEBUG
import html
import re

import requests
from requests.auth import HTTPDigestAuth
from xml.etree import ElementTree as etree

from .exceptions import (
    FritzAuthorizationError,
    FritzConnectionException,
    FRITZ_ERRORS,
)
from .logger import fritzlogger
from .utils import localname


SOAP_NS = "http://schemas.xmlsoap.org/soap/envelope/"
STATUS_UNAUTHORIZED = 401


def datetime_convert(value):
    """
    Converts a string in ISO 8601 format to a datetime-object.
    Raise ValueError if value does not match ISO 8601.
    """
    return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")


def boolean_convert(value):
    """
    Converts a string like '1' or '0' to a boolean value.
    Raise ValueError if it is something else than '1' or '0', because
    this violates the data_type according to the AVM documentation.
    """
    if value == "1" or value == "0":
        return bool(int(value))
    msg = f"value '{value}' does not match '1' or '0'."
    raise ValueError(msg)


def uuid_convert(value):
    """Strips the leading 'uuid:' part from the string."""
    return value.split(":")[-1]


CONVERSION_TABLE = {
    "datetime": datetime_convert,
    "boolean": boolean_convert,
    "uuid": uuid_convert,
    "i4": int,
    "ui1": int,
    "ui2": int,
    "ui4": int,
}


def get_converted_value(data_type, value):
    """
    Try to convert the value from string to the given data_type. The
    data_type is used as key in the CONVERSION_TABLE dictionary. In case
    the data_type is unknown, the original value is returned.
    """
    try:
        return CONVERSION_TABLE[data_type](value)
    except KeyError:
        return value


def encode_boolean(value):
    """
    Returns 1 or 0 if the value is True or False.
    None gets interpreted as False.
    Otherwise, the original value is returned.
    """
    if value is True:
        return 1
    if value is False or value is None:
        return 0
    return value


def get_html_safe_value(value):
    """
    Returns a xml `encoded value` if it's an encodable value.
    `value` can be of any type. If it is a boolean or None it
    gets converted to the integer 1 or 0.
    If it is a string the characters in the set `&<>'"` are
    converted to html-safe sequences.
    Any other datatype gets returned as is.
    """
    value = encode_boolean(value)
    if isinstance(value, str):
        value = html.escape(value)
    return value


def preprocess_arguments(arguments):
    """
    Takes a dictionary with arguments for a soap call and converts all
    values which are True, False or None to the according integers:
    1, 0, 0.
    Returns a new dictionary with the processed values.
    """
    return {k: get_html_safe_value(v) for k, v in arguments.items()}


def get_argument_value(root, argument_name):
    """
    Takes an etree-root object, which is a parsed soap-response from the
    Fritz!Box, and an argument_name, which corresponds to a node-name in
    the element-tree hierarchy. Returns the text-attribute of the node
    as a string.
    Raise an AttributeError in case that a node is not found.
    """
    # root.find will() raise the AttributeError on unknown nodes
    value = root.find(f".//{argument_name}").text
    if value is None:
        # this will be the case on empty tags: <tag></tag>
        value = ""
    return value


def is_html_response(text):
    """
    Returns a boolean whether the raw response text starts with an
    html-tag.
    """
    return text.casefold().startswith("<html")


def remove_html_tags(text):
    """
    Returns the given string `response_text` with all tags removed.
    """
    tag_free = re.sub(r"<.*?>", " ", text)
    return re.sub(r" +", " ", tag_free).strip()  # make it nice


def raise_fritzconnection_error(response):
    """
    Handles all responses with status codes other than 200.
    Will raise a FritzConnectionException with the error code and
    description if available. Can also raise a FritzAuthorizationError
    in case of 401 html-response status code.
    """
    parts = []
    error_code = None

    if is_html_response(response.text):
        # if it is an html response, the error is described in the
        # body part: remove all tags and provide the result as
        # error-message:
        detail = remove_html_tags(response.text)
        msg = f"Unable to perform operation. {detail}"
        if response.status_code == STATUS_UNAUTHORIZED:
            raise FritzAuthorizationError(msg)
        raise FritzConnectionException(msg)

    # otherwise the content is xml and the error-description
    # is part of the structured xml info:
    try:
        root = etree.fromstring(response.content)
    except etree.ParseError as err:
        # should not happen (at least not observed so far)
        raise FritzConnectionException(str(err))

    # extract error information from the provided xml data
    detail = root.find(".//detail")
    children = detail.iter()
    next(children)  # skip detail itself
    for node in children:
        tag = localname(node)
        text = node.text.strip()
        if tag == "errorCode":
            error_code = text
        parts.append(f"{tag}: {text}")
    message = "\n".join(parts)
    # try except:KeyError not possible,
    # because one of the raised Exceptions may inherit from KeyError.
    exception = FRITZ_ERRORS.get(error_code, FritzConnectionException)
    raise exception(message)

def redact_response(redact: bool, input: str):
    # avoid expansive regex matching, when not neccessary
    if fritzlogger.level != DEBUG or not redact:
        return input
    # redact possible phone numbers
    # numbers with len of 5 or more, sourounded by white spaces or bracket
    redacted = re.sub(r"([\s\[\(])\d{5,}([\s\]\)])", r"\1******\2", input)

    # redact external ip addresses
    ext_ip_keys = 'New(ExternalIPAddress|ExternalIPv6Address)'
    redacted = re.sub(r"(<{0}>)(.*)(</{0}>)".format(ext_ip_keys), r"\1******\4", redacted)

    # redact wifi passwords
    wifi_pwd_keys = 'New(WEPKey\d+|PreSharedKey|KeyPassphrase)'
    redacted = re.sub(r"(<{0}>)(.*)(</{0}>)".format(wifi_pwd_keys), r"\1******\4", redacted)
    return redacted

class Soaper:
    """
    Class making the soap on its own to communicate with the FritzBox.
    Instead of ham, spam and eggs, it's hopelessly addicted to soap.

    For accessing the Fritz!Box the parameters `address` for the router
    ip, `port`, `user`, `password` and `session` are required. (These
    parameters will get set by FritzConnection.)
    """

    headers = {"soapaction": "", "content-type": "text/xml", "charset": "utf-8"}

    envelope = re.sub(
        r"\s +",
        "",
        """
        <?xml version="1.0" encoding="utf-8"?>
        <s:Envelope s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"
                    xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">{body}
        </s:Envelope>
        """,
    ).replace('/"xmlns:', '/" xmlns:')

    body_template = re.sub(
        r"\s +",
        "",
        """
        <s:Body>
        <u:{action_name} xmlns:u="{service_type}">{arguments}
        </u:{action_name}>
        </s:Body>
        """,
    )

    argument_template = "<{name}>{value}</{name}>"
    method = "post"

    conversion_table = {
        "datetime": datetime_convert,
        "boolean": boolean_convert,
        "uuid": uuid_convert,
        "i4": int,
        "ui1": int,
        "ui2": int,
        "ui4": int,
    }

    def __init__(self, address, port, user, password, timeout=None, session=None, redact_debug_log=False):
        self.address = address
        self.port = port
        self.user = user
        self.password = password
        self.timeout = timeout
        self.session = session
        self.redact_debug_log = redact_debug_log

    def get_body(self, service, action_name, arguments):
        """Returns the body by template substitution."""
        return self.body_template.format(
            service_type=service.serviceType,
            action_name=action_name,
            arguments=arguments,
        )

    def execute(self, service, action_name, arguments):
        """
        Builds the soap request and returns the response as dictionary.
        Numeric and boolean values are converted from strings to Python
        datatypes.
        """

        def handle_response(response):
            fritzlogger.debug(f"response status: {response.status_code}")
            fritzlogger.debug(redact_response(self.redact_debug_log, response.text))
            if response.status_code != 200:
                raise_fritzconnection_error(response)
            return self.parse_response(response, service, action_name)

        headers = self.headers.copy()
        headers["soapaction"] = f"{service.serviceType}#{action_name}"
        arguments = preprocess_arguments(arguments)
        arguments = "".join(
            self.argument_template.format(name=k, value=v) for k, v in arguments.items()
        )
        body = self.get_body(service, action_name, arguments)
        envelope = self.envelope.format(body=body).encode("utf-8")
        url = f"{self.address}:{self.port}{service.controlURL}"
        fritzlogger.debug(f"\n{url}")
        fritzlogger.debug(envelope)
        if self.session:
            with self.session.post(
                url, data=envelope, headers=headers, timeout=self.timeout
            ) as response:
                return handle_response(response)
        else:
            if self.password:
                auth = HTTPDigestAuth(self.user, self.password)
            else:
                auth = None
            response = requests.post(
                url,
                data=envelope,
                headers=headers,
                auth=auth,
                timeout=self.timeout,
                verify=False,
            )
            return handle_response(response)

    def parse_response(self, response, service, action_name):
        """
        Extracts all known parameters of the given action from the
        response and returns this as a dictionary with the out-parameter
        names as keys and the corresponding response as values.
        Will raise an ActionError on unknown action_name.
        """
        result = dict()
        action = service.actions[action_name]
        root = etree.fromstring(response.content)
        for argument_name in action.arguments:
            try:
                value = get_argument_value(root, argument_name)
            except AttributeError:
                continue
            state_variable_name = action.arguments[argument_name].relatedStateVariable
            state_variable = service.state_variables[state_variable_name]
            data_type = state_variable.dataType.lower()
            try:
                value = get_converted_value(data_type, value)
            except ValueError:
                # ignore malformed value and return 'as is'.
                pass
            result[argument_name] = value
        return result
